module calc_mod
	use fbi_mod
	use secder_mod
	use drlse_mod
	implicit none
	integer, private				:: it, iz, ix, nz, nx, nt, rec_nx, sou_nx, isou
	real, dimension(:,:,:), allocatable, private	:: rec_data, syndata
	real, dimension(:,:), pointer, private 		:: phi_grad

	real, dimension(:,:), allocatable, private	:: residual
	real, dimension(:,:), allocatable, private  	:: phi_a, phi_x, phi_y, laplac_mat, curvature, Nyy, Nxx
	real						:: small_number
	contains





!/////////////////////////////////////////////////////////////////////////////////////
	subroutine calc_init( nz_in, nx_in, nt_in, rec_nx_in, sou_nx_in)
		integer		:: nz_in, nx_in, nt_in, rec_nx_in, sou_nx_in
		nz = nz_in
		nx = nx_in
		nt = nt_in
		rec_nx = rec_nx_in
		sou_nx = sou_nx_in

		small_number = 0.00000000000001

!----------------- allocate --------------------------
		call mem_alloc3d(rec_data,  nt, rec_nx, sou_nx)
		call mem_alloc2d(residual,  nt, rec_nx)
		call mem_alloc3d(syndata,  nt, rec_nx, sou_nx)

		call mem_alloc2d(phi_a, nz, nx)
		call mem_alloc2d(phi_x, nz, nx)
		call mem_alloc2d(phi_y, nz, nx)
		call mem_alloc2d(curvature, nz, nx)
		call mem_alloc2d(laplac_mat, nz, nx)
		call mem_alloc2d(Nxx, nz, nx)
		call mem_alloc2d(Nyy, nz, nx)

!----------------- read in rec_data ------------------
		call sep_read(rec_data, "rec_data")
	end subroutine




!/////////////////////////////////////////////////////////////////////////////////////
	subroutine calc_drlse_term( phi_in, drlse_term, s )
		real, dimension(:,:), target	:: s, phi_in, drlse_term

!----------- Create nabla term (gradient of phi with respect to x and y) ------------
		call neumann_bound_cond(phi_in, phi_a)
		call gradient(phi_a, phi_x, phi_y)
		s = sqrt( phi_x**2 + phi_y**2) ! Nabla term?

!-----------  Calculate DRLSE term  -------------------------------------------------
		Nxx = phi_x/(s + small_number)
		Nyy = phi_y/(s + small_number)

		call div(Nxx,Nyy,curvature)
		call del2(phi_a,laplac_mat)
		drlse_term = 4*laplac_mat - curvature

	end subroutine





!/////////////////////////////////////////////////////////////////////////////////////
	subroutine calc_velocity(phi_in, vel_back, full_vel, Vsalt)
		real				:: Vsalt
		real, dimension(:,:), target	:: phi_in, full_vel, vel_back

		full_vel = vel_back
		do iz=1,nz
			do ix=1,nx
				if (phi_in(iz,ix)>=0.0) then ! This makes assumptions about phi
					full_vel(iz,ix) = Vsalt   ! Inside salt
				end if
			end do
		end do
	end subroutine




!/////////////////////////////////////////////////////////////////////////////////////
	subroutine calc_residual(vel_in, residual2)
		real, dimension(:,:,:), target		:: residual2
		real, dimension(:,:), target		:: vel_in
		residual2 = 0.0
		syndata = 0.0
		call model_update(vel_in)
		call L_nlop2(.false., syndata)
		residual2 = syndata - rec_data
	end subroutine




!/////////////////////////////////////////////////////////////////////////////////////
	subroutine phi_grad_operator(model, data, t1, grad, scaling, adj)
		real, dimension(:,:,:), target			:: data
		real, dimension(:,:), target			:: model, t1, grad, scaling
		logical, intent(in)				:: adj

		t1 = 0.0
		if (adj) then	! residual ==> phi_grad
			model  = 0.0
			call L_lop2(.true. , .false., t1, data)
			model = grad*scaling*t1
		else		! phi_grad ==> residual
			data = 0.0
			t1 = scaling*grad*model
			call L_lop2(.false., .false., t1, data)
		end if
	end subroutine



!/////////////////////////////////////////////////////////////////////////////////////
	subroutine smoother(model, smoothmodel, radius)
		real, dimension(:,:), target	:: model, smoothmodel
		integer				:: radius, left, right, top, bottom, countr, i, j, k, l
		real				:: val

		do j=1, size(model,2)            ! x axis
			do i=1, size(model,1)    ! z axis

				top    = i - radius
				bottom = i + radius
				left   = j - radius
				right  = j + radius

				if (top    < 1) top  = 1
				if (left   < 1) left = 1

				if (bottom > size(model,1)) bottom = size(model,1)
				if (right  > size(model,2)) right  = size(model,2)

				countr = 0
				val = 0.0
				do k=top, bottom
					do l=left, right
						countr = countr + 1
						val = val + model(k,l)
					end do
				end do
				smoothmodel(i,j) = real(val)/real(countr)
!write(0,*) "countr = ", countr, "left = ", left, "right = ", right, "top = ", top, "bottom = ", bottom
!write(0,*) "size(model,1) = ", size(model,1), "size(model,2) = ", size(model,2)
			end do
		end do
	end subroutine



!/////////////////////////////////////////////////////////////////////////////////////
	subroutine dotproduct2(model, data, adj, add)
		real, dimension(:,:), target	:: model
		real				:: data
		logical				:: adj, add
	
		if (adj) then
			model = 0.0
			do iz=1, size(model,1)
				do ix=1, size(model,2)
					model(iz,ix) = model(iz,ix) + data*data
				end do
			end do
		else
			data = 0.0
			do iz=1, size(model,1)
				do ix=1, size(model,2)
					data = data + model(iz,ix)*model(iz,ix)
				end do
			end do
		end if
	end subroutine



!/////////////////////////////////////////////////////////////////////////////////////
	subroutine dotproduct3(model, data, adj, add)
		real, dimension(:,:,:), target	:: model
		real				:: data
		logical				:: adj, add
	
		if (adj) then
			model = 0.0
			do iz=1, size(model,1)
				do ix=1, size(model,2)
					do it=1,size(model,3)
						model(iz,ix,it) = model(iz,ix,it) + data*data
					end do
				end do
			end do
		else
			data = 0.0
			do iz=1, size(model,1)
				do ix=1, size(model,2)
					do it=1,size(model,3)
						data = data + model(iz,ix,it)*model(iz,ix,it)
					end do
				end do
			end do
		end if
	end subroutine



end module
