module fbi_mod

  use sep
  use mem_mod
  use tomo_mod
  use sscale_mod
  use lint_mod
  use step_mod
  use scats0_mod
  use scatr0_mod
  use secder_mod

  implicit none
  integer, private                                     	:: nt, n1, n2, nt2, wfld_j, it
  integer, private                                     	:: sou_nx, sou_ox, sou_dx, sou_z, isou, sou_x
  integer, private                                     	:: rec_nx, rec_ox, rec_dx, rec_z, irec, rec_x, edge_par_shot, edge_par_rec
  integer, private                                     	:: bc1, bc2, bc3, bc4
  real, private                                        	:: dt, d1, o1, d2, o2, dtw
  real, dimension(:,:), allocatable, target, private   	:: veldt2, vwt, wt!,vel, refl,, !Not present in main: veldt2, vwt, wt. ! Not sure refl is necessary
  real, dimension(:,:,:), allocatable, private         	:: dobs
  real, dimension(:,:,:), allocatable, target, private 	:: wflds0, wfldr0
  real, dimension(:,:,:), allocatable, target, private 	:: temp1, temp2, temp1i, temp2i
!--------------------------------------------------------------------------------------
  real, dimension(:), pointer, private      	        :: wavelet ! This is a copy of what is in main. Could use a pointer instead if you want.
!--------------------------------------------------------------------------------------
  contains



  subroutine init1(n1_in, o1_in, d1_in, n2_in, o2_in, d2_in, bc1_in, bc2_in, bc3_in, bc4_in) ! Copying variables 1
    integer                             :: n1_in, n2_in, bc1_in, bc2_in, bc3_in, bc4_in
    real                                :: d1_in, o1_in, d2_in, o2_in
    n1 = n1_in
    n2 = n2_in
    d1 = d1_in
    o1 = o1_in
    d2 = d2_in
    o2 = o2_in
    bc1 = bc1_in
    bc2 = bc2_in
    bc3 = bc3_in
    bc4 = bc4_in
  end subroutine



  subroutine init2(sou_nx_in, sou_ox_in, sou_dx_in, sou_z_in, rec_nx_in, rec_ox_in, rec_dx_in, rec_z_in, edge_par_shot_in, edge_par_rec_in) ! Copying variables 2
    integer                             :: sou_nx_in, sou_ox_in, sou_dx_in, sou_z_in
    integer                             :: rec_nx_in, rec_ox_in, rec_dx_in, rec_z_in, edge_par_shot_in, edge_par_rec_in
    sou_nx = sou_nx_in
    sou_ox = sou_ox_in
    sou_dx = sou_dx_in
    sou_z = sou_z_in
    rec_nx = rec_nx_in
    rec_ox = rec_ox_in
    rec_dx = rec_dx_in
    rec_z = rec_z_in
    edge_par_shot = edge_par_shot_in
    edge_par_rec = edge_par_rec_in
  end subroutine



  subroutine init3(nt_in, dt_in, wfld_j_in) ! Copying variables 3
    integer                             :: nt_in, wfld_j_in
    real                                :: dt_in
    nt = nt_in
    wfld_j = wfld_j_in
    dt = dt_in
  end subroutine



  subroutine init_alloc(wavelet_in) ! Allocations
    real, dimension(nt), target			:: wavelet_in 

    nt2 = (nt-1)*wfld_j+1
    dtw = dt / wfld_j

    call mem_alloc2d(veldt2, n1, n2)
    call mem_alloc2d(vwt, n1, n2)
    call mem_alloc2d(wt, n1, n2)
    call mem_alloc3d(wflds0, n1, n2, nt)
    call mem_alloc3d(wfldr0, n1, n2, nt)
    call mem_alloc3d(dobs, nt, rec_nx, sou_nx)
    call mem_alloc3d(temp1, n1, n2, nt)
    call mem_alloc3d(temp2, n1, n2, nt)
    call mem_alloc3d(temp1i, n1, n2, nt2)
    call mem_alloc3d(temp2i, n1, n2, nt2)

    wavelet => wavelet_in 

    call sscale_init(n1, n2, nt, dtw, veldt2, wt)
    call lint_init(n1, n2, nt, wfld_j)
    call step_init(n1, n2, nt2, bc1, bc2, bc3, bc4, d1, d2, veldt2, vwt, wt)
    call scats0_init(n1, n2, nt, wflds0)
    call scatr0_init(n1, n2, nt, wfldr0)
    call secder_init(n1, n2, nt, dt)

  end subroutine



  subroutine model_update(vel_total)!_in) ! Copy velocity model and prepare boundary condition stuff
    real, dimension(n1, n2)             :: vel_total !_in ,target
    vwt = 0.
    wt = 0.
    call AbcWt(vel_total, vwt, wt, dtw, d1, d2, bc1, bc2, bc3, bc4)
    veldt2 = vel_total*vel_total*dtw*dtw
  end subroutine




	subroutine forward_wavefield(full_wave,rec_wave, isou) ! Nonlinear operator: source => full nonlinear data (equation 9)
		integer					:: isou
		real					:: edge_weight
		real, dimension(:,:,:)			:: full_wave
		real, dimension(nt, rec_nx)		:: rec_wave
		full_wave = 0.0
		rec_wave = 0.0
	!----------------------------------------------------------
		if (isou < edge_par_shot) then
			edge_weight = ( cos( (real(isou-edge_par_shot)/real(edge_par_shot)) * (3.14159265358/2.0) ) )**2
			write(0,*) "edge_weight = ", edge_weight, "isou = ",isou
		else if (isou > sou_nx-edge_par_shot) then
			edge_weight = (cos( (real(isou-sou_nx+edge_par_shot)/real(edge_par_shot))*3.14159265358/2.0))**2
			write(0,*) "edge_weight = ", edge_weight, "isou = ",isou
		else
			edge_weight=1.0
		end if
		write(0,*) "isou = ",isou
	!----------------------------------------------------------
		sou_x = (isou-1)*sou_dx + sou_ox
		temp1 = 0.
		temp1(sou_z,sou_x,:) = wavelet*edge_weight ! K operator
		call sscale_op2(.false., .false., temp1, temp2) ! S operator
		call lint_op2(.false., .false., temp2, temp1i) ! L operator
		call step_op2(.false., .false., temp1i, temp2i) ! G operator
		call lint_op2(.true., .false., temp1, temp2i) ! L' operator
		full_wave= temp1	! temp(z,x,t)
	!----------------------------------------------------------
		do irec=1,rec_nx ! K' operator
			rec_x = (irec-1)*rec_dx + rec_ox
			rec_wave(:,irec) = rec_wave(:,irec) + temp1(rec_z,rec_x,:)
		end do
	!----------------------------------------------------------
	end subroutine



 

  subroutine L_nlop2(add, data) ! Nonlinear operator: source => full nonlinear data (equation 9)
    logical, intent(in)                  :: add
    real, dimension(nt, rec_nx, sou_nx)  :: data
    ! Forward: source => full nonlinear data
    if(.not. add) data = 0.
    do isou=1,sou_nx
      sou_x = (isou-1)*sou_dx + sou_ox
      temp1 = 0.
      temp1(sou_z,sou_x,:) = wavelet ! K operator
      call sscale_op2(.false., .false., temp1, temp2) ! S operator
      call lint_op2(.false., .false., temp2, temp1i) ! L operator
      call step_op2(.false., .false., temp1i, temp2i) ! G operator
      call lint_op2(.true., .false., temp1, temp2i) ! L' operator
!==========================================================================
if (exist_file('full_wave' )) call sep_write(temp1, 'full_wave' )
!==========================================================================
      do irec=1,rec_nx ! K' operator
        rec_x = (irec-1)*rec_dx + rec_ox
        data(:,irec,isou) = data(:,irec,isou) + temp1(rec_z,rec_x,:)
      end do
    end do
  end subroutine








  subroutine L_lop2(adj, add, model, data) ! Born operator: perturbation <=> First order scattered data (i.e. data due to perturbation) (equation 22)
    logical, intent(in)                  :: adj, add
    real, dimension(n1, n2)              :: model
    real, dimension(nt, rec_nx, sou_nx)  :: data
    if(adj) then ! Adjoint: First order scattered data => perturbation
      if(.not. add) model = 0.
      do isou=1,sou_nx
        ! Precompute the u_s0 wavefield using the wavelet
        sou_x = (isou-1)*sou_dx + sou_ox
        temp1 = 0.
        temp1(sou_z,sou_x,:) = wavelet ! K operator
        call sscale_op2(.false., .false., temp1, temp2) ! S operator
        call lint_op2(.false., .false., temp2, temp1i) ! L operator
        call step_op2(.false., .false., temp1i, temp2i) ! G operator
        call lint_op2(.true., .false., wflds0, temp2i) ! L' operator

        ! operator starts here
        temp1 = 0.
        do irec=1,rec_nx ! K operator
          rec_x = (irec-1)*rec_dx + rec_ox
          temp1(rec_z,rec_x,:) = data(:,irec,isou)
        end do
        call lint_op2(.false., .false., temp1, temp2i) ! L operator
        call step_op2(.true., .false., temp1i, temp2i) ! G' operator
        call lint_op2(.true., .false., temp1, temp1i) ! L' operator
        call sscale_op2(.true., .false., temp2, temp1) ! S' operator
        call secder_op2(.true., .false., temp1, temp2) ! T' operator
        call scats0_op2(.true., .true., model, temp1) ! (U_s0)' operator
      end do
    else ! Forward: perturbation => First order scattered data
      if(.not. add) data = 0.
      do isou=1,sou_nx
        ! Precompute the u_s0 wavefield using the wavelet
        sou_x = (isou-1)*sou_dx + sou_ox
        temp1 = 0.
        temp1(sou_z,sou_x,:) = wavelet ! K operator
        call sscale_op2(.false., .false., temp1, temp2) ! S operator
        call lint_op2(.false., .false., temp2, temp1i) ! L operator
        call step_op2(.false., .false., temp1i, temp2i) ! G operator
        call lint_op2(.true., .false., wflds0, temp2i) ! L' operator

        ! operator starts here
        call scats0_op2(.false., .false., model, temp1) ! U_s0 operator
        call secder_op2(.false., .false., temp1, temp2) ! T operator
        call sscale_op2(.false., .false., temp2, temp1) ! S operator
        call lint_op2(.false., .false., temp1, temp1i) ! L operator
        call step_op2(.false., .false., temp1i, temp2i) ! G operator
        call lint_op2(.true., .false., temp1, temp2i) ! L' operator
        do irec=1,rec_nx ! K' operator
          rec_x = (irec-1)*rec_dx + rec_ox
          data(:,irec,isou) = data(:,irec,isou) + temp1(rec_z,rec_x,:)
        end do
      end do
    end if
  end subroutine





end module 
