module initialize
  use sep
  use fbi_mod
  use calc_mod
  use levelset_mod
  use drlse_mod
  implicit none
  contains



subroutine init_main(nz,nx,nt,dz,dx,dt,oz,ox,ot,bc1,bc2,bc3,bc4,wfld_j,sou_nx,sou_dx,sou_ox,rec_nx,rec_dx,rec_ox,rec_z,wavelet,tstep,vstep,niter)
	!----------------------------------------------------------------------------------------------
	integer					:: nt, ntw, nz, nx, wfld_j, niter, smooth_rad
	integer                             	:: sou_nx, sou_ox, sou_dx, sou_z, sou_x, isou
	integer                             	:: rec_nx, rec_ox, rec_dx, rec_z, rec_x, irec, edge_par_shot, edge_par_rec
	integer                            	:: bc1, bc2, bc3, bc4
	real                               	:: dt, dz, oz, dx, ox, ot, tstep, vstep, mu, v1
	integer                                 :: fs,fr,fs_z,fr_z
	real, dimension(:), allocatable     	:: wavelet
	call sep_init()
	!----------------------------------------------------------------------------------------------

	!----------------------------------------------------------------------------------------------
	call from_aux('phi_in', "n1", nz)
	call from_aux('phi_in', "o1", oz)
	call from_aux('phi_in', "d1", dz)
	call from_aux('phi_in', "n2", nx)
	call from_aux('phi_in', "o2", ox)
	call from_aux('phi_in', "d2", dx)
	call from_aux("wavelet", "d1", dt)
	call from_aux("wavelet", "n1", ntw)

	!----------------------------------------------------------------------------------------------
	call from_param("bc1", bc1)
	call from_param("bc2", bc2)
	call from_param("bc3", bc3)
	call from_param("bc4", bc4)
	call from_param("wfld_j", wfld_j)
	call from_param("niter", niter)
	call from_param("nt", nt)
	!call from_param("tstep", tstep)
	!call from_param("vstep", vstep)
	call from_param("edge_par_shot", edge_par_shot)
	call from_param("edge_par_rec", edge_par_rec)
	call from_param("ns", sou_nx, nx      )
	call from_param("fs", fs    , nx/2    ); sou_ox=fs!+1
	call from_param("js", sou_dx, 1       )
	call from_param("nr", rec_nx, sou_nx  )
	call from_param("fr", fr    , nx/2    ); rec_ox=fr!+1
	call from_param("jr", rec_dx, 1       )
	call from_param("fs_z", fs_z); sou_z=fs_z!+1
	call from_param("fr_z", fr_z); rec_z=fr_z!+1
	call from_param("q1", v1)
	!call from_param("mu", mu)
	call from_param("ot", ot)
	call from_param("smooth_rad", smooth_rad)	


	!----------------------------------------------------------------------------------------------
	! Check to make sure that the "nt" from the wavelet is longer than the nt from the parameter (recoding time length)
	! Im not certian, but I am pretty sure that the wavelet must be longer than the recorded data length.
	if(nt > ntw) call erexit("In the parameter file, nt is too small. It is currently smaller than the 'nt' of the wavelet. So either make the wavelet 'nt' bigger, or the parameter file nt smaller.")
	
	!----------------------------------------------------------------------------------------------
	if(sou_ox < 1 .or. (sou_nx-1)*sou_dx+sou_ox > nx) call erexit("sou axis values are not correct")
	if(sou_z < 1 .or. sou_z > nz) call erexit("sou_z value is not correct")
	if(rec_ox < 1 .or. (rec_nx-1)*rec_dx+rec_ox > nx) call erexit("rec axis values are not correct")
	if(rec_z  < 1 .or. rec_z > nz) call erexit("rec_z value is not correct")

	!----------------------------------------------------------------------------------------------
! Do I really need these anymore? 
	call to_history("n1", nt)
	call to_history("o1", ot)
	call to_history("d1", dt)
	call to_history("n2", rec_nx)
	call to_history("o2", rec_ox)
	call to_history("d2", rec_dx)
	call to_history("n3", sou_nx)
	call to_history("o3", sou_ox)
	call to_history("d3", sou_dx)

  	!----------------------------------------------------------------------------------------------
	call mem_alloc1d(wavelet, nt)
	call sep_read(wavelet, "wavelet")

	!----------------------------------------------------------------------------------------------
	call init1(nz, oz, dz, nx, ox, dx, bc1, bc2, bc3, bc4)
	call init2(sou_nx, sou_ox, sou_dx, sou_z, rec_nx, rec_ox, rec_dx, rec_z, edge_par_shot,edge_par_rec)
	call init3(nt, dt, wfld_j)
	call init_alloc(wavelet)

	!----------------------------------------------------------------------------------------------
	call levelset_init(nz, nx, v1, nt, rec_nx, sou_nx, dx, dz, dt, smooth_rad)
	call calc_init( nz, nx, nt, rec_nx, sou_nx)
	call secder_init(nz, nx, nt, dt)
	call drlse_init(nz, nx)
	call newton_init( nz, nx, nt, rec_nx, sou_nx, 0.0000001, 18000, 20, v1)
	!----------------------------------------------------------------------------------------------
	if (exist_file('adj_wave')) then
		call to_history("n1",nz, 'adj_wave')
		call to_history("n2",nx, 'adj_wave')
		call to_history("n3",nt, 'adj_wave')
		call to_history("n4",sou_nx, 'adj_wave')
		call to_history("d1",dz, 'adj_wave')
		call to_history("d2",dx, 'adj_wave')
		call to_history("d3",dt, 'adj_wave')
		call to_history("o1",oz, 'adj_wave')
		call to_history("o2",ox, 'adj_wave')
		call to_history("o3",ot, 'adj_wave')
		call to_history("label1",'z [km]', 'adj_wave')
		call to_history("label2",'x [km]', 'adj_wave')
		call to_history("label3",'time [s]', 'adj_wave')
		call to_history("label4",'source number', 'adj_wave')
	end if


	if (exist_file('full_wave')) then
		call to_history("n1",nz, 'full_wave')
		call to_history("n2",nx, 'full_wave')
		call to_history("n3",nt, 'full_wave')
		call to_history("n4",sou_nx, 'full_wave')
		call to_history("d1",dz, 'full_wave')
		call to_history("d2",dx, 'full_wave')
		call to_history("d3",dt, 'full_wave')
		call to_history("o1",oz, 'full_wave')
		call to_history("o2",ox, 'full_wave')
		call to_history("o3",ot, 'full_wave')
		call to_history("label1",'z [km]', 'full_wave')
		call to_history("label2",'x [km]', 'full_wave')
		call to_history("label3",'time [s]', 'full_wave')
		call to_history("label4",'source number', 'full_wave')
	end if
	
	if (exist_file('s')) then
		call to_history("n1",nz, 's')
		call to_history("n2",nx, 's')
		call to_history("n3",niter, 's')
		call to_history("d1",dz, 's')
		call to_history("d2",dx, 's')
		call to_history("o1",oz, 's')
		call to_history("o2",ox, 's')
		call to_history("label1",'z [km]', 's')
		call to_history("label2",'x [km]', 's')
		call to_history("label3",'iteration', 's')
	end if



	if (exist_file('drlse_term')) then
		call to_history("n1",nz, 'drlse_term')
		call to_history("n2",nx, 'drlse_term')
		call to_history("n3",niter, 'drlse_term')
		call to_history("d1",dz, 'drlse_term')
		call to_history("d2",dx, 'drlse_term')
		call to_history("o1",oz, 'drlse_term')
		call to_history("o2",ox, 'drlse_term')
		call to_history("label1",'z [km]', 'drlse_term')
		call to_history("label2",'x [km]', 'drlse_term')
		call to_history("label3",'iteration', 'drlse_term')
	end if


	if (exist_file('tomodata')) then
   		call to_history("n1",nt, 'tomodata')
		call to_history("d1",dt, 'tomodata')
		call to_history("o1",ot, 'tomodata')
   		call to_history("n2",rec_nx, 'tomodata')
		call to_history("d2",rec_dx, 'tomodata')
		call to_history("o2",rec_ox, 'tomodata')
		call to_history("n3",sou_nx, 'tomodata')
		call to_history("n4",niter, 'tomodata')
		call to_history("label1",'Time [s]', 'tomodata')
		call to_history("label2",'Receiver k', 'tomodata')
		call to_history("label3",'Source k', 'tomodata')
		call to_history("label4",'Levelset Iterations', 'tomodata')
	end if



	if (exist_file('bounddata')) then
   		call to_history("n1",nt, 'bounddata')
		call to_history("d1",dt, 'bounddata')
		call to_history("o1",ot, 'bounddata')
   		call to_history("n2",rec_nx, 'bounddata')
		call to_history("d2",rec_dx, 'bounddata')
		call to_history("o2",rec_ox, 'bounddata')
		call to_history("n3",sou_nx, 'bounddata')
		call to_history("n4",niter, 'bounddata')
		call to_history("label1",'Time [s]', 'bounddata')
		call to_history("label2",'Receiver k', 'bounddata')
		call to_history("label3",'Source k', 'bounddata')
		call to_history("label4",'Levelset Iterations', 'bounddata')
	end if


	if (exist_file('residual')) then
   		call to_history("n1",nt, 'residual')
		call to_history("d1",dt, 'residual')
		call to_history("o1",ot, 'residual')
   		call to_history("n2",rec_nx, 'residual')
		call to_history("d2",rec_dx, 'residual')
		call to_history("o2",rec_ox, 'residual')
		call to_history("n3",sou_nx, 'residual')
		call to_history("n4",niter, 'residual')
		call to_history("label1",'Time [s]', 'residual')
		call to_history("label2",'Receiver k', 'residual')
		call to_history("label3",'Source k', 'residual')
		call to_history("label4",'Iteration', 'residual')
	end if


	if (exist_file('junkdata')) then
   		call to_history("n1",nt, 'junkdata')
		call to_history("d1",dt, 'junkdata')
		call to_history("o1",ot, 'junkdata')
   		call to_history("n2",rec_nx, 'junkdata')
		call to_history("d2",rec_dx, 'junkdata')
		call to_history("o2",rec_ox, 'junkdata')
		call to_history("n3",sou_nx, 'junkdata')
		call to_history("n4",niter, 'junkdata')
		call to_history("label1",'Time [s]', 'junkdata')
		call to_history("label2",'Receiver k', 'junkdata')
		call to_history("label3",'Source k', 'junkdata')
		call to_history("label4",'Levelset Iterations', 'junkdata')
	end if




	if (exist_file('syn_data')) then
		call to_history("n1",nt, 'syn_data')
		call to_history("d1",dt, 'syn_data')
		call to_history("o1",ot, 'syn_data')
		call to_history("n2",rec_nx, 'syn_data')
		call to_history("d2",rec_dx, 'syn_data')
		call to_history("o2",rec_ox, 'syn_data')
		call to_history("n3",sou_nx, 'syn_data')
		call to_history("n4",niter, 'syn_data')
		call to_history("label1",'Time [s]', 'syn_data')
		call to_history("label2",'Receiver [km]', 'syn_data')
		call to_history("label3",'Source k', 'syn_data')
		call to_history("label4",'niter', 'syn_data')
	end if



	if (exist_file('vel_input')) then
		call to_history("n1",nz, 'vel_input')
		call to_history("d1",dz, 'vel_input')
		call to_history("o1",oz, 'vel_input')
		call to_history("n2",nx, 'vel_input')
		call to_history("d2",dx, 'vel_input')
		call to_history("o2",ox, 'vel_input')
		call to_history("n3",niter, 'vel_input')
		call to_history("label1",'z [km]', 'vel_input')
		call to_history("label2",'x [km]', 'vel_input')
		call to_history("label3",'niter', 'vel_input')
	end if



	if (exist_file('vel_back')) then
		call to_history("n1",nz, 'vel_back')
		call to_history("d1",dz, 'vel_back')
		call to_history("o1",oz, 'vel_back')
		call to_history("n2",nx, 'vel_back')
		call to_history("d2",dx, 'vel_back')
		call to_history("o2",ox, 'vel_back')
		call to_history("n3",niter, 'vel_back')
		call to_history("label1",'z [km]', 'vel_back')
		call to_history("label2",'x [km]', 'vel_back')
		call to_history("label3",'niter', 'vel_back')
	end if



	if (exist_file('vel_full')) then
   		call to_history("n1",nz, 'vel_full')
   		call to_history("n2",nx, 'vel_full')
		call to_history("n3",niter, 'vel_full')
		call to_history("d1",dz, 'vel_full')
		call to_history("d2",dx, 'vel_full')
		call to_history("o1",oz, 'vel_full')
		call to_history("o2",ox, 'vel_full')
		call to_history("label1",'z [km]', 'vel_full')
		call to_history("label2",'x [km]', 'vel_full')
		call to_history("label3",'iteration', 'vel_full')
	end if




	if (exist_file('tomograd')) then
		call to_history("n1",nz, 'tomograd')
		call to_history("d1",dz, 'tomograd')
		call to_history("o1",oz, 'tomograd')
		call to_history("n2",nx, 'tomograd')
		call to_history("d2",dx, 'tomograd')
		call to_history("o2",ox, 'tomograd')
		call to_history("n3",niter, 'tomograd')
		call to_history("label1",'z [km]', 'tomograd')
		call to_history("label2",'x [km]', 'tomograd')
		call to_history("label3",'niter' , 'tomograd')
	end if


	if (exist_file('phi_grad')) then
		call to_history("n1",nz, 'phi_grad')
		call to_history("n2",nx, 'phi_grad')
		call to_history("n3",niter, 'phi_grad')
		call to_history("d1",dz, 'phi_grad')
		call to_history("d2",dx, 'phi_grad')
		call to_history("o1",oz, 'phi_grad')
		call to_history("o2",ox, 'phi_grad')
		call to_history("label1",'z [km]', 'phi_grad')
		call to_history("label2",'x [km]', 'phi_grad')
		call to_history("label3",'iteration', 'phi_grad')
	end if


	if (exist_file('param_map' )) then
		call to_history("n3", niter,  'param_map')
	end if



	if (exist_file('objfuncval')) then
		call to_history("n1",niter, 'objfuncval')
		call to_history("d1",1.0, 'objfuncval')
		call to_history("o1",0.0, 'objfuncval')
		call to_history("label1",'iteration', 'objfuncval')
	end if



	if (exist_file('alpha')) then
		call to_history("n1",niter, 'alpha')
		call to_history("d1",1.0, 'alpha')
		call to_history("o1",0.0, 'alpha')
		call to_history("label1",'iteration', 'alpha')
	end if



	if (exist_file('beta')) then
		call to_history("n1",niter, 'beta')
		call to_history("d1",1.0, 'beta')
		call to_history("o1",0.0, 'beta')
		call to_history("label1",'iteration', 'beta')
	end if


	if (exist_file('test-lm')) then
		call to_history("n1",nz, 'test-lm')
		call to_history("d1",dz, 'test-lm')
		call to_history("o1",oz, 'test-lm')
		call to_history("n2",nx, 'test-lm')
		call to_history("d2",dx, 'test-lm')
		call to_history("o2",ox, 'test-lm')
		call to_history("label1",'z [km]', 'test-lm')
		call to_history("label2",'x [km]', 'test-lm')
	end if


	if (exist_file('test-nm')) then
		call to_history("n1",nz, 'test-nm')
		call to_history("d1",dz, 'test-nm')
		call to_history("o1",oz, 'test-nm')
		call to_history("n2",nx, 'test-nm')
		call to_history("d2",dx, 'test-nm')
		call to_history("o2",ox, 'test-nm')
		call to_history("label1",'z [km]', 'test-nm')
		call to_history("label2",'x [km]', 'test-nm')
	end if



	if (exist_file('phi_out')) then
		call to_history("n1"    ,nz         ,'phi_out')
		call to_history("n2"    ,nx         ,'phi_out')
		call to_history("n3"    ,niter+1    ,'phi_out')  ! To account for the first frame being phi0
		call to_history("d1"    ,dz         ,'phi_out')
		call to_history("d2"    ,dx         ,'phi_out')
		call to_history("o1"    ,oz         ,'phi_out')
		call to_history("o2"    ,ox         ,'phi_out')
		call to_history("label1",'z [km]'   ,'phi_out')
		call to_history("label2",'x [km]'   ,'phi_out')
		call to_history("label3",'iteration','phi_out')
	end if
	!----------------------------------------------------------------------------------------------

end subroutine







subroutine init_forward(nz,nx,nt,dz,dx,dt,oz,ox,ot,sou_nx,sou_dx,sou_ox,rec_nx,rec_dx,rec_ox,rec_z,wavelet)
	!----------------------------------------------------------------------------------------------
	integer					:: nt, ntw, nz, nx, wfld_j
	integer                             	:: sou_nx, sou_ox, sou_dx, sou_z, sou_x, isou
	integer                             	:: rec_nx, rec_ox, rec_dx, rec_z, rec_x, irec, edge_par_shot, edge_par_rec
	integer                            	:: bc1, bc2, bc3, bc4
	real                               	:: dt, dz, oz, dx, ox, ot
	integer                                 :: fs,fr,fs_z,fr_z
	real, dimension(:), allocatable     	:: wavelet
	real, dimension(:,:), allocatable	:: vel
	!----------------------------------------------------------------------------------------------
	call sep_init()
	!----------------------------------------------------------------------------------------------
	call from_history("n1", nz)
	call from_history("o1", oz)
	call from_history("d1", dz)
	call from_history("n2", nx)
	call from_history("o2", ox)
	call from_history("d2", dx)
	!----------------------------------------------------------------------------------------------
	call from_param("bc1", bc1)
	call from_param("bc2", bc2)
	call from_param("bc3", bc3)
	call from_param("bc4", bc4)
	call from_param("wfld_j", wfld_j)

	call from_param("nt", nt)
	call from_aux("wavelet", "d1", dt)

	call from_param("edge_par_shot", edge_par_shot)
	call from_param("edge_par_rec", edge_par_rec)

	call from_param("ns", sou_nx, nx      )
	call from_param("fs", fs    , nx/2    ); sou_ox=fs!+1
	call from_param("js", sou_dx, 1       )
	call from_param("nr", rec_nx, sou_nx  )
	call from_param("fr", fr    , nx/2    ); rec_ox=fr!+1
	call from_param("jr", rec_dx, 1       )

	call from_param("fs_z", fs_z); sou_z=fs_z!+1
	call from_param("fr_z", fr_z); rec_z=fr_z!+1

	!----------------------------------------------------------------------------------------------
	! Check to make sure that the "nt" from the wavelet is longer than the nt from the parameter (recoding time length)
	! Im not certian, but I am pretty sure that the wavelet must be longer than the recorded data length.
	call from_aux("wavelet", "n1", ntw)
	if(nt > ntw) call erexit("In the parameter file, nt is too small. It is currently smaller than the 'nt' of the wavelet. So either make the wavelet 'nt' bigger, or the parameter file nt smaller.")
	!----------------------------------------------------------------------------------------------
	call from_param("ot", ot)	
	!----------------------------------------------------------------------------------------------
	if(sou_ox < 1 .or. (sou_nx-1)*sou_dx+sou_ox > nx) call erexit("sou axis values are not correct")
	if(sou_z < 1 .or. sou_z > nz) call erexit("sou_z value is not correct")
	if(rec_ox < 1 .or. (rec_nx-1)*rec_dx+rec_ox > nx) call erexit("rec axis values are not correct")
	if(rec_z  < 1 .or. rec_z > nz) call erexit("rec_z value is not correct")
	!----------------------------------------------------------------------------------------------
	call mem_alloc1d(wavelet, nt)
	call sep_read(wavelet, "wavelet")
	!----------------------------------------------------------------------------------------------
	call init1(nz, oz, dz, nx, ox, dx, bc1, bc2, bc3, bc4)
	call init2(sou_nx, sou_ox, sou_dx, sou_z, rec_nx, rec_ox, rec_dx, rec_z, edge_par_shot,edge_par_rec)
	call init3(nt, dt, wfld_j)
	call init_alloc(wavelet)
	!----------------------------------------------------------------------------------------------
	call to_history("n1", nt)
	call to_history("d1", dt)
	call to_history("o1", ot)
	call to_history("n2", rec_nx)
	call to_history("d2", rec_dx)
	call to_history("o2", rec_ox)
	call to_history("n3", sou_nx)
	call to_history("d3", sou_dx)
	call to_history("o3", sou_ox)
	call to_history("label1",'Time [s]')
	call to_history("label2",'Receiver [km]')


	!----------------------------------------------------------------------------------------------
	if (exist_file('full_wave_stack')) then
		call to_history("n1", nz, 'full_wave_stack')
		call to_history("n2", nx, 'full_wave_stack')
		call to_history("n3", nt, 'full_wave_stack')
		call to_history("d1", dz, 'full_wave_stack')
		call to_history("d2", dx, 'full_wave_stack')
		call to_history("d3", dt, 'full_wave_stack')
		call to_history("o1", oz, 'full_wave_stack')
		call to_history("o2", ox, 'full_wave_stack')
		call to_history("o3", ot, 'full_wave_stack')
	end if
	!----------------------------------------------------------------------------------------------
	call mem_alloc2d(vel, nz, nx)
	call sep_read(vel)
	call model_update(vel)
end subroutine






end module
