module newton_mod
	use calc_mod
	use fbi_mod
	use planesearch_mod
	use solver_mod
	use cgstep_mod
	implicit none
	integer, private				:: nz, nx, nt, rec_nx, sou_nx, imax, ntest, ntotal
	real, private					:: sigma, maxbeta, Vsalt, alphaNL, betaNL
	real, dimension(:), pointer, private 		:: data
	real, dimension(:,:), pointer, private 		:: phi_grad, velmodel, vel_back_in, phi_in, bound_grad, tomo_grad
	real, dimension(:,:,:), pointer, private	:: tomodata, bounddata, residual
	real, dimension(:), allocatable, private	:: model, model0
	real, dimension(:,:), allocatable, private	:: tvel, tphi, tvel_back
	real, dimension(:,:,:), allocatable, private	:: tresidual
	contains




!/////////////////////////////////////////////////////////////////////////////////////
	subroutine newton_init(nz_in, nx_in, nt_in, rec_nx_in, sou_nx_in, sigma_in, imax_in, ntestin, Vsalt_in)
		integer					:: nz_in, nx_in, nt_in, rec_nx_in, sou_nx_in, imax_in, ntestin
		real					:: sigma_in, Vsalt_in

		nz       = nz_in
		nx       = nx_in
		nt       = nt_in
		rec_nx   = rec_nx_in
		sou_nx   = sou_nx_in
		sigma    = sigma_in
		imax     = imax_in
		ntest    = ntestin
		ntotal   = nt*rec_nx*sou_nx
		Vsalt    = Vsalt_in

		allocate( tvel(nz,nx), tphi(nz,nx), tvel_back(nz,nx) )
		allocate( tresidual(nt,rec_nx,sou_nx), residual(nt,rec_nx,sou_nx) )
		allocate( model0(2), model(2), data(ntotal) )
		
	end subroutine





!/////////////////////////////////////////////////////////////////////////////////////
	subroutine NLsearch_init(vel_back_in_A, phi_in_A, bound_grad_A, tomo_grad_A, maxbeta_in, alphaNL_in, betaNL_in)
		real, dimension(:,:), target	:: vel_back_in_A, phi_in_A, bound_grad_A, tomo_grad_A
		real				:: Vsalt_in, alphaNL_in, betaNL_in, maxbeta_in
		alphaNL = alphaNL_in
		betaNL = betaNL_in
		vel_back_in  => vel_back_in_A
		phi_in       => phi_in_A
		bound_grad   => bound_grad_A
		tomo_grad    => tomo_grad_A
		maxbeta = maxbeta_in
	end subroutine




!/////////////////////////////////////////////////////////////////////////////////////
	subroutine NL_linesearch(bestgamma)
		real		:: dg, bestgamma, bestobj, objfuncval, gamma, maxgamma
		integer		:: k

		maxgamma = maxbeta/betaNL
		dg = 2*maxgamma/(ntest-1)
		gamma = -1.0*maxgamma
		!dg = maxgamma/(ntest-1)
		!gamma = 0.0

		! Initialize
		tphi = phi_in
		tvel_back = vel_back_in
		call calc_velocity(tphi, tvel_back, tvel, Vsalt)
		call calc_residual(tvel, tresidual )
		call dotproduct3(tresidual, objfuncval, .false., .false.)
		bestobj   = objfuncval*2
		bestgamma = 0.0

		! Loop through the test values
		do k=1,ntest
			tvel=0.0
			tresidual=0.0
			gamma = gamma + real(k-1)*dg
			tphi = phi_in + gamma*betaNL*bound_grad
			tvel_back = vel_back_in + gamma*alphaNL*tomo_grad

			call calc_velocity(tphi, tvel_back, tvel, Vsalt)
			call calc_residual(tvel, tresidual )
			call dotproduct3(tresidual, objfuncval, .false., .false.)
			! Keep if the new obj value is good
			if (objfuncval < bestobj) then
				bestgamma = gamma
				bestobj = objfuncval
			end if
		end do
	end subroutine






!/////////////////////////////////////////////////////////////////////////////////////
	subroutine Lsearch_init(tomodata_in, bounddata_in, residual_in)
		real, dimension(:,:,:), target		:: tomodata_in, bounddata_in
		real, dimension(nt,rec_nx,sou_nx)	:: residual_in

		tomodata   => tomodata_in
		bounddata  => bounddata_in
		residual   = residual_in
		data = reshape(residual, (/(nt*rec_nx*sou_nx)/))

		call planesearch_init(tomodata, bounddata, nt, rec_nx, sou_nx)
	end subroutine





!/////////////////////////////////////////////////////////////////////////////////////
	subroutine L_planesearch( alphaL, betaL )
		real	:: alphaL, betaL, mean1, mean2, mean3

		mean1 = sum(abs(tomodata))/(max(1,size(tomodata)))
		mean2 = sum(abs(bounddata))/(max(1,size(bounddata)))
		mean3 = sum(abs(residual))/(max(1,size(residual)))

		alphaL = mean3/mean1
		betaL  = mean3/mean2
		model0(1) = alphaL
		model0(2) = betaL
		model = 0.0

		call solver_tiny( model, data, planesearch_op, cgstep, 200, model0)
		alphaL = model(1)
		betaL = model(2)
	end subroutine



end module
